#!/usr/bin/env python

import asyncio
import time

from ruia import Request
from ruia.exceptions import InvalidRequestMethod

sem = asyncio.Semaphore(3)


async def hello(response):
    return "hello ruia"


def hi(response):
    yield "hi ruia"


async def retry_func(request):
    request.request_config["TIMEOUT"] = 10


params = {"name": "ruia"}


async def valid_response(response):
    return response


async def make_get_request(sem, callback=None):
    request_config = {
        "RETRIES": 3,
        "DELAY": 1,
        "TIMEOUT": 0.1,
        "VALID": valid_response,
        "RETRY_FUNC": retry_func,
    }
    request = Request(
        "https://httpbin.org/get",
        method="GET",
        metadata={"hello": "ruia"},
        request_config=request_config,
        params=params,
        callback=callback,
    )
    return await request.fetch_callback(sem)


async def make_post_request(sem, callback):
    headers = {"Content-Type": "application/json"}
    request = Request(
        "https://httpbin.org/post",
        method="POST",
        headers=headers,
        data=params,
        callback=callback,
    )
    return await request.fetch_callback(sem)


def test_request_config():
    assert str(Request("https://httpbin.org/")) == "<GET https://httpbin.org/>"
    _, response = asyncio.get_event_loop().run_until_complete(
        make_get_request(sem=sem, callback=hello)
    )
    # assert response.callback_result == "hello ruia"
    assert response.metadata == {"hello": "ruia"}
    json_result = asyncio.get_event_loop().run_until_complete(response.json())
    assert json_result["args"]["name"] == "ruia"

    _, response = asyncio.get_event_loop().run_until_complete(
        make_post_request(sem=sem, callback=None)
    )
    json_result = asyncio.get_event_loop().run_until_complete(response.json())
    assert json_result["data"] == "name=ruia"


def test_method_error_request():
    try:
        request = Request("https://httpbin.org/", method="PUT")
        response = asyncio.get_event_loop().run_until_complete(request.fetch())
        res = asyncio.get_event_loop().run_until_complete(response.text())
        assert res == ""
    except Exception as e:
        assert isinstance(e, InvalidRequestMethod)


def test_sem_error_request():
    _, response = asyncio.get_event_loop().run_until_complete(
        make_get_request(sem=None, callback=None)
    )
    assert response == None


def test_retry_request():
    request = Request("http://httpbin.org/404")
    _, response = asyncio.get_event_loop().run_until_complete(
        request.fetch_callback(sem=sem)
    )
    assert response.url == "http://httpbin.org/404"


def test_retry_delay():
    # Test invalid URL (to trigger retries) with 1s delay between retries
    request_config = {"RETRIES": 2, "RETRY_DELAY": 1}
    request = Request("http://127.0.0.1:5999/", request_config=request_config)

    # Start a timer to time retries
    timer = time.time()
    _, response = asyncio.get_event_loop().run_until_complete(
        request.fetch_callback(sem=sem)
    )

    # Ensure that for 2 retries the time taken is > 2s (1s between each retry)
    assert time.time() - timer > 2


def test_delay_false():
    request_config = {"DELAY": 10}
    request = Request("https://httpbin.org/", request_config=request_config)

    # Start a timer to time request
    timer = time.time()
    response = asyncio.get_event_loop().run_until_complete(request.fetch(delay=False))

    # Ensure delay option was ignored (time taken is less than 10s)
    assert time.time() - timer < 10


def test_timeout_request():
    async def timeout_request(sem):
        request_config = {"RETRIES": 1, "DELAY": 1, "TIMEOUT": 0.1}
        request = Request(
            "https://httpbin.org/get",
            method="GET",
            metadata={"hello": "ruia"},
            encoding="utf-8",
            request_config=request_config,
            params=params,
            callback=hi,
        )
        return await request.fetch_callback(sem)

    _, response = asyncio.get_event_loop().run_until_complete(timeout_request(sem=sem))
    assert "https://httpbin.org/get" in response.url



#!/usr/bin/env python
"""
 Created by howie.hu at 2019/1/28.
"""
import asyncio

from lxml import etree
from ruia import Request


async def hello(response):
    return "hello ruia"


sem = asyncio.Semaphore(3)
params = {"name": "ruia"}
request = Request(
    "https://httpbin.org/get",
    method="GET",
    metadata={"hello": "ruia"},
    params=params,
    callback=hello,
)
_, response = asyncio.get_event_loop().run_until_complete(request.fetch_callback(sem))


def test_response():
    url = response.url
    method = response.method
    encoding = response.encoding
    metadata = response.metadata
    cookies = response.cookies
    history = response.history
    headers = response.headers
    status = response.status

    text = asyncio.get_event_loop().run_until_complete(response.text())
    json = asyncio.get_event_loop().run_until_complete(response.json())
    read = asyncio.get_event_loop().run_until_complete(response.read())

    assert url == "https://httpbin.org/get?name=ruia"
    assert method == "GET"
    assert encoding == "utf-8"
    assert metadata == {"hello": "ruia"}
    assert isinstance(cookies, dict)
    assert history == ()
    assert headers["Content-Type"] == "application/json"
    assert status == 200
    assert response.ok == True
    assert isinstance(response.html_etree(html=text), etree._Element)
    assert isinstance(text, str)
    assert isinstance(json, dict)
    assert isinstance(read, bytes)

    assert (
        str(response)
        == "<Response url[GET]: https://httpbin.org/get?name=ruia status:200>"
    )


def test_callback():
    assert response.callback_result is None
    response.callback_result = "ruia"
    assert response.callback_result == "ruia"


def test_index():
    assert response.index is None
    response.index = "ruia"
    assert response.index == "ruia"


def test_ok():
    assert response.ok == True
    response.ok = False
    assert response.ok == False



#!/usr/bin/env python

import asyncio
import os

from ruia import AttrField, Item, TextField
from ruia.exceptions import IgnoreThisItem, InvalidFuncType

html_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "data", "for_item_testing.html"
)
with open(html_path, mode="r", encoding="utf-8") as file:
    HTML = file.read()


class DoubanItems(Item):
    target_item = TextField(css_select="div.item")
    title = TextField(css_select="span.title")
    cover = AttrField(css_select="div.pic>a>img", attr="src")
    abstract = TextField(css_select="span.inq")

    async def clean_title(self, title):
        if isinstance(title, str):
            return title
        else:
            return "".join([i.text.strip().replace("\xa0", "") for i in title])


class DoubanItem(Item):
    title = TextField(css_select="head title")
    constant_attr = "hello ruia"

    async def clean_title(self, title):
        return "Title: " + title


class DoubanCleanMethodErrorItem(Item):
    title = TextField(css_select="head title")

    def clean_title(self, title):
        return "Title: " + title


class DoubanIgnoreItem(Item):
    title = TextField(css_select="head title")

    async def clean_title(self, title):
        raise IgnoreThisItem


class HackerNewsItem(Item):
    target_item = TextField(css_select="tr.athing")
    title = TextField(css_select="a.storylink")
    url = AttrField(css_select="a.storylink", attr="href")


async def parse_item(html):
    items = []
    async for item in DoubanItems.get_items(html=html):
        items.append(item)
    return items


async def error_parse_item(html):
    items = []
    async for item in DoubanItem.get_items(html=html):
        items.append(item)
    return items


async def single_page_demo(url="https://news.ycombinator.com/"):
    items = []
    async for item in HackerNewsItem.get_items(url=url, sem=asyncio.Semaphore(3)):
        items.append(item)
    return items


def test_item():
    item = asyncio.get_event_loop().run_until_complete(DoubanItem.get_item(html=HTML))
    assert item.title == "Title: 豆瓣电影TOP250"
    assert item.constant_attr == "hello ruia"

    try:
        asyncio.get_event_loop().run_until_complete(
            DoubanCleanMethodErrorItem.get_item(html=HTML)
        )
    except Exception as e:
        assert isinstance(e, InvalidFuncType)

    try:

        asyncio.get_event_loop().run_until_complete(DoubanItem.get_item(html=""))
    except Exception as e:
        assert isinstance(e, ValueError)

    try:
        asyncio.get_event_loop().run_until_complete(
            DoubanItem.get_item(html_etree="test")
        )
    except Exception as e:
        assert isinstance(e, AttributeError)


def test_items():
    items = asyncio.get_event_loop().run_until_complete(parse_item(html=HTML))
    assert items[0].abstract == "希望让人自由。"

    try:
        asyncio.get_event_loop().run_until_complete(error_parse_item(html=HTML))
    except Exception as e:
        assert isinstance(e, ValueError)


def test_item_results():
    item = asyncio.get_event_loop().run_until_complete(DoubanItem.get_item(html=HTML))
    assert item.results == {"title": "Title: 豆瓣电影TOP250"}
    assert item.ignore_item == False


def test_ignore_item():
    item = asyncio.get_event_loop().run_until_complete(
        DoubanIgnoreItem.get_item(html=HTML)
    )
    assert item.ignore_item == True


def test_items_results():
    items = asyncio.get_event_loop().run_until_complete(parse_item(html=HTML))
    assert items[0].results["title"] == "肖申克的救赎"


def test_request_url():
    items = asyncio.get_event_loop().run_until_complete(single_page_demo())
    assert isinstance(items, list)
    assert len(items) > 1
    assert "Item" in str(items[0])


def test_fail_on_both_html_url():
    try:
        item = asyncio.get_event_loop().run_until_complete(
            DoubanIgnoreItem.get_item(html=HTML, url="testing")
        )
    except Exception as e:
        assert type(e) == ValueError



#!/usr/bin/env python

import asyncio
import os

from ruia import Item, Middleware, Request, Response, Spider, TextField
from ruia.exceptions import SpiderHookError

html_path = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "data", "for_spider_testing.html"
)
with open(html_path, mode="r", encoding="utf-8") as file:
    HTML = file.read()

middleware = Middleware()


async def retry_func(request):
    request.request_config["TIMEOUT"] = 10


@middleware.request
async def print_on_request(spider_ins, request):
    request.headers = {"User-Agent": "ruia ua"}


@middleware.response
async def print_on_response(spider_ins, request, response):
    assert isinstance(await response.text(), str)
    assert request.headers == {"User-Agent": "ruia ua"}


class ItemDemo(Item):
    title = TextField(xpath_select="/html/head/title")


class SpiderDemo(Spider):
    start_urls = ["https://httpbin.org/get?p=0"]
    request_config = {"RETRIES": 3, "DELAY": 0, "TIMEOUT": 20}
    headers = {"User-Agent": "Ruia Spider"}
    aiohttp_kwargs = {}

    call_nums = 0

    async def parse(self, response):
        yield Request(
            url=response.url,
            callback=self.parse_item,
            headers=self.headers,
            request_config=self.request_config,
            **self.aiohttp_kwargs,
        )

    async def parse_item(self, response):
        pages = [f"https://httpbin.org/get?p={i}" for i in range(1, 2)]
        async for resp in self.multiple_request(pages):
            yield self.parse_next(response=resp, any_param="hello")

    async def parse_next(self, response, any_param):
        assert any_param == "hello"
        yield self.request(
            url="https://httpbin.org/get?p=2",
            metadata={"any_param": any_param},
            callback=self.parse_details,
        )

    async def parse_details(self, response):
        item = await ItemDemo.get_item(html=HTML)
        yield item

    async def process_item(self, item: ItemDemo):
        assert item.title == "for_spider_testing"
        await self.count_nums()

    async def count_nums(self):
        SpiderDemo.call_nums += 1


def test_spider_with_middleware():
    loop = asyncio.new_event_loop()
    SpiderDemo.start(loop=loop, middleware=middleware)
    assert SpiderDemo.call_nums == 1


def test_spider_with_error_middleware():
    error_middleware = Middleware()

    @error_middleware.request
    def error_request(spider_ins, request, response):
        pass

    @error_middleware.response
    async def error_response(spider_ins, request, response):
        raise TypeError("error")

    class SpiderDemo(Spider):
        start_urls = ["https://httpbin.org/get?p=0"]

        async def parse(self, response):
            pass

    SpiderDemo.start(middleware=error_middleware)


def test_spider_hook():
    async def after_start_func(spider_ins):
        print("after_start_func")
        spider_ins.result["after_start"] = True
        assert isinstance(spider_ins.result, dict)

    async def before_stop_func(spider_ins):
        print("before_stop_func")
        spider_ins.result["before_stop"] = True

    class SpiderHook(Spider):
        start_urls = ["https://httpbin.org/get?p=0", "https://httpbin.org/404"]
        request_config = {"RETRIES": 1, "DELAY": 0, "TIMEOUT": 10}

        result = {
            "after_start": False,
            "before_stop": False,
            "process_succeed_response": False,
            "process_failed_response": False,
            "process_item": False,
        }

        async def parse(self, response):
            item = await ItemDemo.get_item(html=HTML)
            yield item

        async def process_item(self, item):
            self.result["process_item"] = True

        async def process_succeed_response(self, request, response):
            # Hook for response
            self.result["process_succeed_response"] = True

        async def process_failed_response(self, request, response):
            # Hook for response
            self.result["process_failed_response"] = True

    # Test middleware & hook
    loop = asyncio.new_event_loop()
    SpiderHook.start(
        loop=loop, after_start=after_start_func, before_stop=before_stop_func
    )

    assert SpiderHook.result["after_start"] == True
    assert SpiderHook.result["before_stop"] == True
    assert SpiderHook.result["process_succeed_response"] == True
    assert SpiderHook.result["process_failed_response"] == True
    assert SpiderHook.result["process_item"] == True


def test_spider_hook_error():
    class SpiderDemo(Spider):
        start_urls = ["https://httpbin.org/get?p=0"]

        async def parse(self, response):
            pass

    async def before_stop_func(spider_ins):
        raise TypeError("error")

    loop = asyncio.new_event_loop()
    try:
        SpiderDemo.start(loop=loop, before_stop=before_stop_func)
    except Exception as e:
        assert isinstance(e, SpiderHookError)


def test_invalid_callback_result():
    class SpiderDemo(Spider):
        start_urls = ["https://httpbin.org/get?p=0"]
        result = {"process_callback_result": False}

        async def parse(self, response):
            yield {}

    async def process_dict_callback_result(spider_ins, callback_result):
        spider_ins.result["process_callback_result"] = True

    class CustomCallbackResultType:
        @classmethod
        def init_spider(cls, spider):
            spider.callback_result_map = spider.callback_result_map or {}
            setattr(
                spider, "process_dict_callback_result", process_dict_callback_result
            )
            spider.callback_result_map.update({"dict": "process_dict_callback_result"})

    CustomCallbackResultType.init_spider(SpiderDemo)

    loop = asyncio.new_event_loop()
    SpiderDemo.start(loop=loop)
    assert SpiderDemo.result["process_callback_result"] == True


def test_spider_multiple_request_sync():
    result = list()

    class MultipleRequestSpider(Spider):
        start_urls = ["https://httpbin.org"]
        concurrency = 3

        async def parse(self, response: Response):
            urls = [f"https://httpbin.org/get?p={page}" for page in range(1, 2)]
            async for response in self.multiple_request(urls, is_gather=True):
                yield self.parse_next(response=response)

        async def parse_next(self, response):
            json_result = await response.json()
            page = json_result["args"]["p"]
            result.append(int(page))

    MultipleRequestSpider.start()
    assert result == [1]


def test_no_start_url_spider():
    try:

        class NoStartUrlSpider(Spider):
            pass

        NoStartUrlSpider.start()
    except Exception as e:
        assert isinstance(e, ValueError)


def test_callback_error():
    class NoParseSpider(Spider):
        start_urls = ["https://httpbin.org/get"]

    NoParseSpider.start()

    class CallbackError(Spider):
        start_urls = ["https://httpbin.org/get"]

        async def parse(self, response):
            raise ValueError("error")

    CallbackError.start()


def test_coroutine_callback_error():
    class CoroutineItemErrorSpider(Spider):
        start_urls = ["https://httpbin.org/get"]

        async def parse(self, response):
            pages = ["https://httpbin.org/get?p=1"]
            async for resp in self.multiple_request(pages):
                yield self.parse_item(response=resp)

        async def parse_item(self, response):
            await ItemDemo.get_item(html=await response.text())

    CoroutineItemErrorSpider.start()

    class CoroutineErrorSpider(Spider):
        start_urls = ["https://httpbin.org/get"]

        async def parse(self, response):
            pages = ["https://httpbin.org/get?p=1"]
            async for resp in self.multiple_request(pages):
                yield self.parse_item(response=resp)

        async def parse_item(self, response):
            raise ValueError("error")

    CoroutineErrorSpider.start()


def test_nothing_matched_spider():
    class NothingMatchedErrorSpider(Spider):
        start_urls = ["https://httpbin.org/get"]

        async def parse(self, response):
            await ItemDemo.get_item(html=await response.text())

    NothingMatchedErrorSpider.start()


def test_multiple_spider():
    class MultipleSpider(Spider):
        count = 0
        start_urls = ["https://httpbin.org/get?p=0"]

        async def parse(self, response):
            MultipleSpider.count += 1

    async def multiple_spider(loop):
        await MultipleSpider.async_start(loop=loop, middleware=[middleware])
        await MultipleSpider.async_start(loop=loop, middleware=middleware)
        return MultipleSpider

    loop = asyncio.new_event_loop()
    spider_ins = loop.run_until_complete(multiple_spider(loop=loop))
    assert spider_ins.count == 2
